'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
from webuntis import utils, objects, errors
import re

try:
    import urllib.request as urlrequest
    import urllib.error as urlerrors
except ImportError:
    import urllib2 as urlrequest
    import urllib2 as urlerrors
import logging
import datetime


try:
    import json
except ImportError:
    import simplejson as json


class JSONRPCSession(object):
    '''Contains the functions for not much more than a simple JSON-RPC Session.
    Can be used as a context-handler.
    '''
    options = None

    def __init__(self, **kwargs):
        # The OptionStore is an extended dictionary, associating validators
        # and other helper methods with each key
        self.options = utils.FilterDict(utils.option_utils.option_parsers)
        options = {
            'server': None,
            'school': None,
            'useragent': None,
            'username': None,
            'password': None,
            'jsessionid': None,
            'cachelen': 20
        }
        options.update(kwargs)

        self.options.update({
            'server': options['server'],
            'school': options['school'],
            'useragent': options['useragent'],
            'credentials': {
                'username': options['username'],
                'password': options['password'],
                'jsessionid': options['jsessionid']
            }
        })
        if options['cachelen'] > 0:
            self._cache = utils.LruDict(maxlen=options['cachelen'])

    def __enter__(self):
        '''Context-manager'''
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''Context-manager -- the only thing we need to clean up is to log out
        '''
        self.logout()

    def logout(self, suppress_errors=False):
        '''
        Log out of session

        :param suppress_errors: boolean, whether to not raise an error if we
            already were logged out.
        '''
        #Send a JSON-RPC 'logout' method without parameters to log out
        if 'jsessionid' in self.options['credentials']:
            self._make_request('logout')
            del self.options['credentials']['jsessionid']
        elif not suppress_errors:
            raise errors.RemoteError('Already logged out!')

    def login(self):
        '''Initializes an authentication, provided we have the credentials for
        it.

        .. note::

            When authenticating, the username and password will be deleted from
            memory and only the jsessionid will be kept in the
            :py:class:`webuntis.utils.option_utils.OptionStore` instance.
        '''

        if 'username' not in self.options['credentials'] \
                or 'password' not in self.options['credentials']:
            raise errors.AuthError('No login data specified.')

        logging.debug('Trying to authenticate with username/password...')
        logging.debug('Username: ' +
                      self.options['credentials']['username'] +
                      ' Password: ' +
                      self.options['credentials']['password'])
        res = self._make_request('authenticate', {
            'user': self.options['credentials']['username'],
            'password': self.options['credentials']['password'],
            'client': self.options['useragent']
        })
        logging.debug(res)
        if 'sessionId' in res:
            logging.debug('Did get a jsessionid from the server:')
            self.options['credentials']['jsessionid'] = res['sessionId']
            logging.debug(self.options['credentials']['jsessionid'])
        else:
            raise errors.AuthError('Something went wrong while \
                    authenticating', res)

        return self


    def _request(self, method, params=None):
        '''A wrapper for _make_request implementing a LRU Cache'''
        key = (method, hash(tuple(params or {})))
        
        if key not in self._cache:
            self._cache[key] = self._make_request(method, params)
        return self._cache[key]

    def _make_request(self, method, params=None):
        '''
        A method for sending a JSON-RPC request.

        :param method: The JSON-RPC method to be executed
        :type method: str

        :param params: JSON-RPC parameters to the method \
                (should be JSON serializable)
        :type params: dict
        '''

        url = self.options['server']
        url += '?school=' + self.options['school']
        cookie_header = True
        if method == 'authenticate':
            cookie_header = False
        elif 'jsessionid' not in self.options['credentials']:
            raise errors.AuthError(
                'Don\'t have JSESSIONID. Did you already log out?'
            )

        if not params:
            params = {}

        req_data = {
            'id': str(datetime.datetime.today()),
            'method': method,
            'params': params,
            'jsonrpc': '2.0'
        }

        req_data_json = json.dumps(req_data).encode()

        logging.debug('Making new request:')
        logging.debug('URL: ' + url)
        logging.debug(req_data_json)

        req = urlrequest.Request(
            url,
            req_data_json,
            {
                'User-Agent': self.options['useragent'],
                'Content-Type': 'application/json'
            }
        )
        if cookie_header:
            req.add_header(
                'Cookie',
                'JSESSIONID=' + self.options['credentials']['jsessionid']
            )

        # this will eventually raise errors, e.g. if there's an unexpected http
        # status code
        res = urlrequest.urlopen(req)

        res_str = res.read().decode('utf-8')

        try:
            res_data = json.loads(res_str)
            logging.debug('Valid JSON found')
            logging.debug(res_data)
        except ValueError as e:
            logging.error('Error: Invalid JSON')
            logging.debug(res_str)
            raise errors.RemoteError('Invalid JSON')

        if res_data['id'] != req_data['id']:
            raise errors.RemoteError('Request id was not the same as \
                the one returned')
        elif 'result' in res_data:
            return res_data['result']
        else:
            logging.error(res_data)
            raise errors.RemoteError(
                'Some JSON-RPC-ish error happened',
                str(res_data),
                str(req_data)
            )


class Session(JSONRPCSession):
    '''Provides an abstraction layer above the JSON-RPC Instance.
    '''

    def __getattr__(self, name):
        '''Returns a callable which creates an instance (or reuses an old one)
        of the appropriate object-list class
        '''
        def get_objectlist(**kwargs):
            '''Returns the corrent objectclass after checking in the cache
            '''
            return objects.object_lists[name](session=self, kwargs=kwargs)

        if name in objects.object_lists:
            return get_objectlist
        else:
            raise AttributeError('There is no such datatype')



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals

import itertools

def table(timetable):
    '''
    Creates a table-like nested list out of an instance of
    :py:class:`webuntis.objects.periods.PeriodList`.

    Returns a list containing tuples of "rows" and a datetime object, which in
    turn contain tuples of "hours" and their weekday number, which
    contain :py:class:`webuntis.objects.periods.PeriodObject` instances which
    are happening at the same time.

    >>> from webuntis.utils.timetable_utils import table
    >>> table(s.timetable(klasse=109))
    '''

    if not len(timetable):
        return []

    time = lambda x: int(x.start.strftime('%H%M%S'))
    date = lambda x: int(x.start.strftime('%Y%m%d'))

    grouped = []
    times = []
    for i, row in itertools.groupby(sorted(timetable, key=time), time):
        new_row = []

        starttime = None

        # group lessons by date -- now all lessons occuring at the same time
        # are grouped
        for i2, hour in itertools.groupby(sorted(row, key=date), date):
            if hour:
                # itertools._grouper objects can't be accessed like lists
                hour = list(hour)
                weekday = hour[0].start.weekday()

                while (len(new_row) - 1) < weekday:
                    new_row.append(list())  # expand row as much as needed
                new_row[weekday] = hour

                if not starttime:
                    starttime = hour[0].start

        grouped.append(new_row)
        times.append(starttime)

    # expand each row to the maximal length
    longest_row = len(max(grouped, key=len))
    for row in grouped:
        while len(row) < longest_row:
            row.append(list())

    # at last add a weekday indicator
    return zip(times, [enumerate(row) for row in grouped])



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals

from . import datetime_utils, option_utils

from collections import OrderedDict


class lazyproperty(object):
    '''A read-only @property that is only evaluated once.

    Stolen from http://www.reddit.com/r/Python/comments/ejp25/cached_property_decorator_that_is_memory_friendly/
    '''
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class LruDict(OrderedDict):
    def __init__(self, maxlen=50):
        super(LruDict, self).__init__()
        self._maxlen = maxlen

    def __setitem__(self, key, value):
        super(LruDict, self).__setitem__(key, value)
        while len(self.items()) > self._maxlen:
            self.popitem(last=False)


class FilterDict(dict):
    '''A dictionary which passes new values to a function found at the
    corresponding key in self.filters

    :param filters: A dictionary containing functions. If a new key is set into
    an instance of FilterDict, the filter dictionary will be consulted to
    filter the value with a function.

    >>> options = FilterDict({
    ...    'foo': lambda x: 'whoopdeedoo'
    ... })
    >>>
    >>> options['foo'] = 'somethingelse'
    >>> options['foo']
    'whoopdeedoo'

    '''
    filters = None

    def __init__(self, filters):
        self.filters = filters

    def __getitem__(self, name):
        # check if we got a real Option subclass
        if name in self and dict.__getitem__(self, name):
            # every Option subclass has this
            return dict.__getitem__(self, name)
        elif name in self.filters:
            raise KeyError('No value for key: {}'.format(name))
        else:
            raise KeyError('No value or filter for key: {}'.format(name))

    def __setitem__(self, key, value):
        new_value = self.filters[key](value)
        if new_value:
            dict.__setitem__(
                self,
                key,
                new_value
            )
        elif key in self:
            dict.__delitem__(self, key)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self.__setitem__(key, value)



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals

from webuntis.utils import datetime_utils, lazyproperty


class Result(object):
    '''Class used to represent most API method results.
    '''
    _jsonrpc_method = False
    _data = False

    def __init__(self, session, kwargs):
        if not self._jsonrpc_method:
            raise NotImplementedError

        self._session = session
        self._data = self._get_data(**kwargs)

    def _jsonrpc_parameters(self, **kwargs):
        '''This method returns all methods that should be passed to the
        JSON-RPC request'''
        # we return an empty dictionary here so subclasses don't
        # really need to overwrite this if they actually have no parameters
        return {}

    def _get_data(self, **kwargs):
        '''A simple wrapper for the jsonrpc_parameters builder.
        Can be overwritten by subclasses, which may require more complex
        options'''
        return self._session._request(
            self._jsonrpc_method,
            self._jsonrpc_parameters(**kwargs)
        )


class ListItem(object):
    '''ListItems represent an item in a
    :py:class:`webuntis.objects.ListResult`. They don\'t contain methods to
    retrieve data.'''

    #: the raw JSON data returned from the server
    _data = None

    #: the ID of this element. When dealing with arrays as result, it is very
    #  common for a item to have its own ID.
    id = None

    def __init__(self, session, parent, data):
        '''
        Keyword arguments:
        session -- a webuntis.session.Session object
        data -- the relevant part of a json result for this object
        '''
        self._session = session
        self._parent = parent
        self._data = data

        self.id = self._data['id'] if 'id' in self._data else None

    def __int__(self):
        '''This is useful if the users pass a ListItem when a numerical ID
        is expected, so we just can put the thing through int(), regardless of
        what type it is.'''
        return self.id


class ListResult(Result):
    '''A list-like version of *Result* that takes a list and returns a list of
    objects, containing a list value each.
    '''

    # When the Result returns an array, this is very useful. Every item of that
    # array will be fed to an instance of self._itemclass, with the session and
    # the array item as initialization arguments.

    #: the class which should be used to instantiate an array item.
    _itemclass = ListItem

    def filter(self, **criterions):
        '''
        Returns a list of all objects, filtered by attributes::

            foo = s.klassen().filter(id=1)  # returns the same as...
            foo = [kl for kl in s.klassen() if kl.id == 1]

            bar = s.klassen().filter(id=(1,2,3,4))  # returns the same as...
            bar = [kl for kl in s.klassen() if kl.id in (1,2,3,4)]

        ::
        '''
        def meets_criterions(item):
            '''Returns true if the item meets the criterions'''
            for key, value in criterions.items():
                # if the attribute value isn't one we're looking for
                if getattr(item, key) in value:
                    continue
                elif getattr(item, key) == value:
                    continue
                else:
                    return False

            return True

        return [item for item in self if meets_criterions(item)]

    def __getitem__(self, i):
        '''Makes the object iterable and behave like a list'''
        if not isinstance(self._data[i], ListItem):
            # if we don't have an object yet
            self._data[i] = self._itemclass(self._session, self, self._data[i])

        return self._data[i]

    def __len__(self):
        '''Return the length of the items'''
        return len(self._data)


class DepartmentObject(ListItem):
    '''Represents a department
    '''

    @lazyproperty
    def name(self):
        '''short name such as *R1A*'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name, such as *Raum Erste A*. Not predictable.'''
        return self._data['longName']


class DepartmentList(ListResult):
    '''A list of all departments::

        departments = s.departments()

    ::
    '''
    _itemclass = DepartmentObject
    _jsonrpc_method = 'getDepartments'


class HolidayObject(ListItem):
    '''Represents a single holiday.
    '''

    @lazyproperty
    def start(self):
        '''The start date of the holiday, as a datetime object.'''
        return datetime_utils.parse_date(self._data['startDate'])

    @lazyproperty
    def end(self):
        '''The end of the holiday'''
        return datetime_utils.parse_date(self._data['endDate'])

    @lazyproperty
    def name(self):
        '''Name, such as *Nationalfeiertag*.'''
        return self._data['longName']

    @lazyproperty
    def short_name(self):
        '''Abbreviated form of the name'''
        return self._data['name']


class HolidayList(ListResult):
    '''A list of all holidays::

        s.holidays()

    ::
    '''
    _itemclass = HolidayObject
    _jsonrpc_method = 'getHolidays'


class KlassenObject(ListItem):
    '''Represents a school class.'''

    @lazyproperty
    def name(self):
        '''Name of class'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name of class'''
        return self._data['longName']


class KlassenList(ListResult):
    '''A list of all school classes.

    :param schoolyear:
        The schoolyear where we should get all our school \
        classes from. Either a \
        :py:class:`webuntis.objects.SchoolyearObject` or an \
        id of it.

    ::
        >>> s.klassen()
        >>>
        >>> year = s.schoolyears().filter(id=2)
        >>> s.klassen(schoolyear=year)

    '''
    _itemclass = KlassenObject
    _jsonrpc_method = 'getKlassen'

    def _jsonrpc_parameters(self, schoolyear=None):
        jsonrpc_parameters = {}
        if schoolyear:
            jsonrpc_parameters.update({
                'schoolyearId': int(schoolyear)
            })

        return jsonrpc_parameters


class PeriodObject(ListItem):
    '''Represents a time range, where lessons/subjects may be held.
    '''

    @lazyproperty
    def start(self):
        '''The start date/time of the period, as datetime object.'''

        return datetime_utils.parse_datetime(
            self._data['date'],
            self._data['startTime']
        )

    @lazyproperty
    def end(self):
        '''The end date/time of the period.'''

        return datetime_utils.parse_datetime(
            self._data['date'],
            self._data['endTime']
        )

    @lazyproperty
    def klassen(self):
        '''A list of :py:class:`webuntis.objects.KlassenObject` instances,
        which are attending this period.'''

        return self._session.klassen().filter(
            id=[kl['id'] for kl in self._data['kl']]
        )

    @lazyproperty
    def teachers(self):
        '''A list of :py:class:`webuntis.objects.TeacherObject` instances,
        which are attending this period.'''

        return self._session.teachers().filter(
            id=[te['id'] for te in self._data['te']]
        )

    @lazyproperty
    def subjects(self):
        '''A list of :py:class:`webuntis.objects.SubjectObject` instances,
        which are topic of this period. This is not used for things like
        multiple language lessons (*e.g.* Latin, Spanish, French) -- each of
        those will get placed in their own period.'''

        return self._session.subjects().filter(
            id=[su['id'] for su in self._data['su']]
        )

    @lazyproperty
    def rooms(self):
        '''The rooms where this period is taking place at. This also is not
        used for multiple lessons, but rather for a single lesson that is
        actually occuring at multiple locations.'''

        return self._session.rooms().filter(
            id=[ro['id'] for ro in self._data['ro']]
        )

    @lazyproperty
    def type(self):
        '''May be "cancelled" or "irregular" -- the latter implies a
        substitution.'''

        return self._data['lstype'] if 'lstype' in self._data else None


class PeriodList(ListResult):
    '''
    Aka timetable

    :param start: a starting date either in the form YYMMDD or as date/time
        object.
    :param end: a ending date in the same form as the starting date.

    Furthermore you have to explicitly define a klasse, teacher, subject, room
    or student parameter containing the id or the object of the thing you want
    to get a timetable about::

        schoolclass = s.klassen().filter(id=1)  # schoolclass #1

        s.timetable(klasse=schoolclass)  # which is the same as...
        s.periods(klasse=schoolclass)

    .. note:: See :py:mod:`webuntis.utils.timetable_utils` for various tools
        around the timetable.
    '''
    _itemclass = PeriodObject
    _jsonrpc_method = 'getTimetable'

    def _jsonrpc_parameters(self, start=None, end=None, **type_and_id):
        element_type_table = {
            'klasse':  1,
            'teacher': 2,
            'subject': 3,
            'room':    4,
            'student': 5
        }

        if len(type_and_id) != 1:
            raise ValueError(
                'You have to specify exactly one of the following parameters \
                    by keyword: ' +
                ', '.join(element_type_table.keys())
            )

        # apply end to start and vice-versa if one of them is missing
        if not start and end:
            start = end
        elif not end and start:
            end = start

        element_type, element_id = list(type_and_id.items())[0]

        # if we have to deal with an object in element_id,
        # its id gets placed here anyway
        parameters = {
            'id': int(element_id),
            'type': element_type_table[element_type],
        }

        if start:
            parameters['startDate'] = datetime_utils.format_date(start)
        if end:
            parameters['endDate'] = datetime_utils.format_date(end)

        return parameters


class RoomObject(ListItem):
    '''Represents a physical room. Such as a classroom, but also the physics
    laboratory or whatever.
    '''
    @lazyproperty
    def name(self):
        '''The short name of the room. Such as PHY.'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''The long name of the room. Such as "Physics lab".'''
        return self._data['longName']


class RoomList(ListResult):
    '''
    Represents a list of rooms::

        s.rooms()

    '''
    _itemclass = RoomObject
    _jsonrpc_method = 'getRooms'


class SchoolyearObject(ListItem):
    '''Represents a schoolyear.
    '''

    @lazyproperty
    def name(self):
        '''"2010/2011"'''

        return self._data['name']

    @lazyproperty
    def start(self):
        '''The start date of the schoolyear, as datetime object'''
        return datetime_utils.parse_date(self._data['startDate'])

    @lazyproperty
    def end(self):
        '''The end date'''
        return datetime_utils.parse_date(self._data['endDate'])

    @lazyproperty
    def is_current(self):
        '''
        Boolean, check if this is the current schoolyear::

            >>> y = s.schoolyears()
            >>> y.current.id
            7
            >>> y.current.is_current
            True
            >>> y.filter(id=y.current.id).is_current
            True

        '''
        return (self == self._parent.current)


class SchoolyearList(ListResult):
    '''
    Represents a list of school years::

        s.schoolyears()

    '''
    _itemclass = SchoolyearObject
    _jsonrpc_method = 'getSchoolyears'

    @lazyproperty
    def current(self):
        '''Returns the current schoolyear in form of a
        :py:class:`webuntis.objects.SchoolyearObject`'''
        current_data = self._session._request('getCurrentSchoolyear')
        current = self.filter(id=current_data['id'])[0]
        return current


class SubjectObject(ListItem):
    '''Represents a subject.
    '''

    @lazyproperty
    def name(self):
        '''Short name of subject, such as *PHY*'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name of subject, such as *Physics*'''
        return self._data['longName']


class SubjectList(ListResult):
    '''Represents a list of subjects::

        s.subjects()

    ::
    '''
    _itemclass = SubjectObject
    _jsonrpc_method = 'getSubjects'


class TeacherObject(ListItem):
    '''Represents a teacher.
    '''
    @lazyproperty
    def fore_name(self):
        '''fore name of the teacher'''
        return self._data['foreName']

    @lazyproperty
    def long_name(self):
        '''surname of teacher'''
        return self._data['longName']

    surname = long_name

    @lazyproperty
    def name(self):
        '''full name of the teacher'''
        return self._data['name']


class TeacherList(ListResult):
    '''
    Represents a list of teachers::

        s.teachers()  # get all teachers of school

    ::
    '''
    _itemclass = TeacherObject
    _jsonrpc_method = 'getTeachers'


class TimeunitObject(ListItem):
    '''A bunch of timeunits for a specific day.
    '''

    @lazyproperty
    def times(self):
        '''A list of tuples containing the start and the end of each timeunit
        as datetime '''

        return [
            (
                datetime_utils.parse_time(unit['startTime']),
                datetime_utils.parse_time(unit['endTime'])
            ) for unit in self._data['timeUnits']
        ]

    @lazyproperty
    def day(self):
        '''The day the timeunit list is for'''
        return self._data['day']


class TimeunitList(ListResult):
    '''A list of times and dates for the current week. Doesn't contain actual
    data, but is useful when you want to generate a timetable::

        >>> grid = s.timegrid()
        >>>
        >>> # 1 = Sunday
        >>> # 2 = Monday
        >>> # ...
        >>> # 7 = Saturday
        >>> grid[0].day
        2
        >>> grid[0].times
        [
            (datetime.datetime(...), datetime.datetime(...)),
            ...
        ]

    .. note::
        The date properties of the datetime objects are invalid! Since these
        are not provided by the official API, there's not much you can do about
        it.
    '''

    _itemclass = TimeunitObject
    _jsonrpc_method = 'getTimegridUnits'


class ColorInfo(object):
    '''
    An object containing information about a lession type or a period code::

        >>> lstype = s.statusdata().lession_types[ls]
        >>> lstype.name
        'ls'
        >>> lstype.forecolor
        '000000'
        >>> lstype.backcolor
        'ee7f00'

    '''

    def __init__(self, session, parent, data):
        self._session = session
        self._parent = parent
        self._data = data

    @lazyproperty
    def name(self):
        '''The name of the LessionType or PeriodCode'''
        return list(self._data.items())[0][0]

    @lazyproperty
    def forecolor(self):
        '''The foreground color used in the web interfacei and elsewhere'''
        return self._data[self.name]['foreColor']

    @lazyproperty
    def backcolor(self):
        '''The background color used in the web interface and elsewhere'''
        return self._data[self.name]['backColor']


class StatusData(Result):
    '''
    Information about lession types and period codes and their colors::

        s.statusdata()

    '''
    _jsonrpc_method = 'getStatusData'

    @lazyproperty
    def lession_types(self):
        '''A list of :py:class:`webuntis.objects.ColorInfo` objects, containing
        information about all lession types defined'''
        return [
            ColorInfo(self._session, self, data)
            for data in self._data['lstypes']
        ]

    @lazyproperty
    def period_codes(self):
        '''A list of :py:class:`webuntis.objects.ColorInfo` objects, containing
        information about all period codes defined'''
        return [
            ColorInfo(self._session, self, data)
            for data in self._data['codes']
        ]

# Defines result classes that are accessible from outside
object_lists = {
    'departments': DepartmentList,
    'holidays': HolidayList,
    'klassen': KlassenList,
    'rooms': RoomList,
    'schoolyears': SchoolyearList,
    'subjects': SubjectList,
    'teachers': TeacherList,

    'timegrid': TimeunitList,
    'timeunits': TimeunitList,
    'statusdata': StatusData,

    'timetable': PeriodList,
    'periods': PeriodList
}



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
import re
import logging
try:
    from urllib import parse as urlparse
except ImportError:
    import urlparse


def careless_parser(value):
    return value


def credentials_parser(creds):
    if 'username' in creds and 'password' in creds \
       and creds['username'] and creds['password']:
        return {
            'username': creds['username'],
            'password': creds['password']
        }
    elif 'jsessionid' in creds and creds['jsessionid']:
        return {
            creds['jsessionid']
        }
    else:
        return {}


def server_parser(value):
    if not value:
        return value  # either it's None or we have a dictionary

    if re.match(r'^http(s?)\:\/\/', value):
        url = value
    else:  # if we just have the domain
        logging.debug('The URL given doesn\'t seem to be a valid URL, \
            just gonna prepend "http://"')

        # append the http prefix and hope for the best
        url = 'http://' + value

    urlobj = urlparse.urlparse(url)

    if urlobj.scheme and urlobj.netloc:
        if urlobj.path == '/':
            # A bit weird, but it formats kinda nicely in the log.
            # I am not logging line each, because other threads might
            # interfere this and fuck everything up.
            logging.warning('''You specified that the API endpoint
should be /. That is uncommon. If you didn't mean to do so, remove the slash at
the end of your "server" parameter.''')

        return urlobj.scheme + \
            '://' + \
            urlobj.netloc + \
            (urlobj.path or '/WebUntis/jsonrpc')
    else:
        return None  # Doesn't seem like we have a value.

option_parsers = {
    'credentials': credentials_parser,
    'school': careless_parser,
    'server': server_parser,
    'useragent': careless_parser
}



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
from datetime import datetime as datetime


def parse_date(string):
    return _parse(string, '%Y%m%d')


def parse_time(string):
    return _parse(string, '%H%M')


def parse_datetime(date, time):
    return _parse(str(date) + '///' + str(time), '%Y%m%d///%H%M')


def _parse(string, formatting):
    if type(string) == int:
        string = str(string)

    return datetime.strptime(string, formatting)


def format_date(obj):
    return _format(obj, '%Y%m%d')


def format_time(obj):
    return _format(obj, '%H%M')


def _format(obj, formatting):
    return int(datetime.strftime(obj, formatting))



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals


class Error(Exception):
    pass


class RemoteError(Error, IOError):
    pass


class AuthError(RemoteError):
    pass



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
import unittest
import mock
import os
import sys
import webuntis
from webuntis.utils import timetable_utils, datetime_utils, option_utils
import webuntis.utils as utils
import json

class WebUntisTests:
    tests_path = os.path.abspath(os.path.dirname(__file__))
    data_path = tests_path + '/static'


class WebUntisOfflineTests(unittest.TestCase, WebUntisTests):
    def setUp(self):
        self.request_patcher = patcher = mock.patch(
            'webuntis.Session._request',
            side_effect=Exception('This is an OFFLINE testsuite!')
        )
        patcher.start()

        self.session = webuntis.Session()

    def tearDown(self):
        self.request_patcher.stop()
        self.session = None

    def test_credentials(self):
        self.assertRaises(KeyError, self.session.options.__getitem__, ('credentials'))

    def test_options_invalidattribute(self):
        try:
            self.session.options['nigglywiggly']
        except KeyError:
            pass
        else:
            raise Exception('Expected that OptionStore fails with invalid \
                attribute')

    def test_options_no_attribute(self):
        self.assertEqual(self.session.options, {})

    def test_getdepartments_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getdepartments_mock.json')
        )

        with mock.patch.object(
            webuntis.objects.DepartmentList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for dep_raw, dep in zip(jsonstr, self.session.departments()):
                self.assertEqual(dep_raw['id'], dep.id)
                self.assertEqual(dep_raw['longName'], dep.long_name)
                self.assertEqual(dep_raw['name'], dep.name)

    def test_getholidays_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getholidays_mock.json')
        )

        with mock.patch.object(
            webuntis.objects.HolidayList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for holiday_raw, holiday in zip(jsonstr, self.session.holidays()):
                self.assertEqual(
                    holiday_raw['startDate'],
                    int(holiday.start.strftime('%Y%m%d'))
                )
                self.assertEqual(
                    holiday_raw['endDate'],
                    int(holiday.end.strftime('%Y%m%d'))
                )
                self.assertEqual(
                    holiday_raw['id'],
                    holiday.id
                )
                self.assertEqual(
                    holiday_raw['name'],
                    holiday.short_name
                )
                self.assertEqual(
                    holiday_raw['longName'],
                    holiday.name
                )

    def test_getklassen_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getklassen_mock.json')
        )

        with mock.patch.object(
            webuntis.objects.KlassenList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for klasse_raw, klasse in zip(jsonstr, self.session.klassen()):
                self.assertEqual(klasse_raw['longName'], klasse.long_name)
                self.assertEqual(klasse_raw['name'], klasse.name)
                self.assertEqual(klasse_raw['id'], klasse.id)

    def test_gettimetables_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/gettimetables_mock.json')
        )
        periodlist_class = webuntis.objects.PeriodList

        with mock.patch.object(
            periodlist_class,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            tt = self.session.timetable(klasse=114)

        for period_raw, period in zip(jsonstr, tt):
            self.assertEqual(
                int(period.start.strftime('%H%M')),
                period_raw['startTime']
            )
            self.assertEqual(
                int(period.start.strftime('%Y%m%d')),
                period_raw['date']
            )

            self.assertEqual(
                int(period.end.strftime('%H%M')),
                period_raw['endTime']
            )
            self.assertEqual(
                int(period.end.strftime('%H%M')),
                period_raw['endTime']
            )

    def test_getrooms_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getrooms_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.RoomList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for room_raw, room in zip(jsonstr, self.session.rooms()):
                self.assertEqual(room_raw['longName'], room.long_name)
                self.assertEqual(room_raw['name'], room.name)
                self.assertEqual(room_raw['id'], room.id)

    def test_getschoolyears_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getschoolyears_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.SchoolyearList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for year_raw, year in zip(jsonstr, self.session.schoolyears()):
                self.assertEqual(
                    year_raw['startDate'],
                    int(year.start.strftime('%Y%m%d'))
                )
                self.assertEqual(
                    year_raw['endDate'],
                    int(year.end.strftime('%Y%m%d'))
                )
                self.assertEqual(year_raw['name'], year.name)
                self.assertEqual(year_raw['id'], year.id)

    def test_getsubjects_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getsubjects_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.SubjectList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for subj_raw, subj in zip(jsonstr, self.session.subjects()):
                self.assertEqual(subj_raw['id'], subj.id)
                self.assertEqual(subj_raw['longName'], subj.long_name)
                self.assertEqual(subj_raw['name'], subj.name)

    def test_getteachers_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getteachers_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.TeacherList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for t_raw, t in zip(jsonstr, self.session.teachers()):
                self.assertEqual(t_raw['longName'], t.long_name)
                self.assertEqual(t_raw['longName'], t.surname)
                self.assertEqual(t_raw['foreName'], t.fore_name)
                self.assertEqual(t_raw['name'], t.name)


    def test_gettimegrid_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/gettimegrid_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.TimeunitList,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            for t_raw, t in zip(jsonstr, self.session.timegrid()):
                self.assertEqual(t_raw['day'], t.day)
                for t2_raw, t2 in zip(t_raw['timeUnits'], t.times):
                    self.assertEqual(t2_raw['startTime'],
                                     int(t2[0].strftime('%H%M')))
                    self.assertEqual(t2_raw['endTime'],
                                     int(t2[1].strftime('%H%M')))

    def test_getstatusdata_mock(self):
        jsonstr = json.load(
            open(self.data_path + '/getstatusdata_mock.json')
        )
        with mock.patch.object(
            webuntis.objects.StatusData,
            '_get_data',
            return_value=jsonstr
        ) as test_mock:
            statusdata = self.session.statusdata()
            for lstype_raw, lstype in zip(jsonstr['lstypes'],
                                          statusdata.lession_types):
                name = list(lstype_raw.items())[0][0]
                colors = lstype_raw[name]
                self.assertEqual(name, lstype.name)
                self.assertEqual(colors['foreColor'], lstype.forecolor)
                self.assertEqual(colors['backColor'], lstype.backcolor)

    def test_filterdict(self):
        store = utils.FilterDict({
            'whatever': lambda x: x,
            'always_whoop': lambda x: 'whoop'
        })
        store['whatever'] = 'lel'
        self.assertEqual(store['whatever'], 'lel')
        
        store['always_whoop'] = 'what'
        self.assertEqual(store['always_whoop'], 'whoop')

        del store['whatever']
        self.assertRaises(KeyError, store.__getitem__, ('whatever'))

        del store['always_whoop']
        self.assertRaises(KeyError, store.__getitem__, ('always_whoop'))


class WebUntisRemoteTests(unittest.TestCase, WebUntisTests):
    def setUp(self):
        self.session = webuntis.session.Session(
            school='demo_inf',
            server='webuntis.grupet.at:8080',
            username='api',
            password='api',
            useragent='WebUntis Unittests'
        ).login()
        print('\n' + '*' * 100)

    def tearDown(self):
        try:
            self.session.logout()
        except:
            logging.warning('Was not able to log out!')
        finally:
            self.session = None

    def test_timetableutils_table_emptyinput(self):
        if len(webuntis.utils.timetable_utils.table([])):
            raise Exception
    ## OBJECTS

    def test_getdepartments(self):
        self.session.departments()

    def test_getholidays(self):
        self.session.holidays()

    def test_getklassen(self):
        print('TEST_GETKLASSEN')
        klassen = self.session.klassen()

        for klasse in klassen:
            self.assertEqual(klasse.__class__,
                             webuntis.objects.KlassenObject)

            klasse.id
            klasse.name
            klasse.long_name

    def test_getrooms(self):
        self.session.rooms()

    def test_getschoolyears(self):
        self.session.schoolyears()

    def test_getsubjects(self):
        self.session.subjects()

    def test_getteachers(self):
        print('TEST_GETTEACHERS')
        teachers = self.session.teachers()

        for teacher in teachers:
            self.assertEqual(teacher.__class__,
                             webuntis.objects.TeacherObject)

            teacher.fore_name
            teacher.name
            teacher.surname
            teacher.long_name
            teacher.id

    def test_gettimeunits(self):
        self.session.timeunits()

    def test_gettimetables(self):
        print('TEST_GETTIMETABLES')
        timetables = []
        for i, klasse in enumerate(self.session.klassen()):
            if i > 5:
                break
            timetable = self.session.timetable(klasse=klasse)
            print('Length of timetable: ' + str(len(timetable)))
            timetables.append(timetable)

    ## /OBJECTS

    def test_emptyfilter(self):
        print('TEST_EMPTYFILTER')
        for klasse, klasse_filtered, i in zip(
            self.session.klassen(),
            self.session.klassen().filter(),
            range(len(self.session.klassen()))
        ):
            if klasse.id != klasse_filtered.id:
                raise Exception(
                    'test_emptyfilter failed: On position no. ' +
                    str(i) + ' there are not the same two klassen'
                )

    def test_loggedout(self):
        self.session.logout()
        creds = self.session.options['credentials']
        if 'jsessionid' in creds:
            raise Exception('Expected to be logged out, \
                but this is what i\'ve got in creds: ' + str(creds))

        self.assertRaises(webuntis.errors.AuthError, self.session.klassen)


def main():
    try:
        raw_input  # check if we have python2
    except NameError:
        prompt = input
    else:
        prompt = raw_input

    def load(test):
        suite = unittest.TestLoader().loadTestsFromTestCase(test)
        unittest.TextTestRunner(verbosity=2).run(suite)

    tests = [WebUntisOfflineTests, WebUntisRemoteTests]

    if len(sys.argv) == 1:
        print('ERROR: No tests specified.\n\nAdd a number to the command-line arguments to execute the test:')
        for i, test in enumerate(tests):
            print('#{}: {}'.format(i, test.__name__))
        exit(1)

    checked_tests = [tests[int(x)] for x in sys.argv[1:]]

    for test in checked_tests:
        prompt('Hit enter to continue.')
        load(test)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from setuptools import setup
from sys import version_info

if version_info < (2,6):
    dependencies = ['simplejson']
else:
    dependencies = []

setup(
    name='webuntis',
    version='0.1.0',
    author='Markus Unterwaditzer',
    author_email='markus@unterwaditzer.net',
    packages=['webuntis', 'webuntis.utils', 'webuntis.tests'],
    include_package_data=True,
    url='http://dev.unterwaditzer.net/python-webuntis/',
    license='new-style BSD',
    description='Bindings for WebUntis API',
    long_description=open('README.rst').read(),
    install_requires=dependencies
)
